/* libguestfs
 * Copyright (C) 2017 Red Hat Inc.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

/**
 * Utility functions using libxml2.
 *
 * These functions these I<must not> call internal library functions
 * such as C<safe_*>, C<error> or C<perrorf>, or any C<guestfs_int_*>.
 */

#include <config.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <locale.h>

#include <libxml/uri.h>

#include "c-ctype.h"

/* NB: MUST NOT include "guestfs-internal.h". */
#include "guestfs.h"
#include "guestfs-utils.h"
#include "libxml2-utils.h"

/**
 * This is a wrapper around C<xmlParseURI>.  That function cannot
 * handle spaces and some non-ASCII characters found in URIs.  This
 * wrapper URI-encodes those before calling C<xmlParseURI> and returns
 * the URI structure.
 *
 * This function should B<only> be called for the URIs that libguestfs
 * has invented, for things like guestfish I<--add> and virt-v2v.
 *
 * For real URIs or libvirt URIs this may cause corruption in corner
 * cases.  (See L<https://news.ycombinator.com/item?id=11673058>
 * describing some of the complexity involved in dealing with real
 * URI).
 *
 * On error, returns C<NULL> and sets C<errno> appropriately.
 *
 * Caller must call C<xmlFreeURI> on the returned structure or use the
 * C<CLEANUP_XMLFREEURI> cleanup macro.
 */
xmlURIPtr
guestfs_int_parse_nonstandard_uri (const char *arg)
{
  CLEANUP_FREE char *uri = NULL;
  CLEANUP_FREE char *escaped_uri = NULL;
  static const char hexdigit[] = "0123456789abcdef";
  size_t i, j, len;
  xmlURIPtr ret;

  /* Convert the string to UTF-8. */
  uri = guestfs_int_local_string_to_utf8 ((char *) arg);
  if (uri == NULL)
    return NULL;

  /* Since we know the URI is in well-formed UTF-8 we can iterate over
   * the bytes to do the escaping.  The output of this will never be
   * more than 3 times larger (each byte might be rewritten as ‘%XX’).
   */
  len = strlen (uri);
  escaped_uri = malloc (3*len + 1);
  if (escaped_uri == NULL)
    return NULL;

  for (i = j = 0; i < strlen (uri); ++i) {
    /* See RFC 3986 appendix A.  Note this leaves existing %-encoded
     * escapes alone.
     */
    if (c_isalnum (uri[i]) ||
        strchr ("%-._~:/?#[]@!$&'()*+,;=", uri[i]) != NULL)
      escaped_uri[j++] = uri[i];
    else {
      escaped_uri[j++] = '%';
      escaped_uri[j++] = hexdigit [(((unsigned char) uri[i]) >> 4) & 0xf];
      escaped_uri[j++] = hexdigit [((unsigned char) uri[i]) & 0xf];
    }
  }
  escaped_uri[j++] = '\0';

  /* libxml2 xmlParseURI does not reliably set errno, so it's likely
   * best to ignore whatever errno is returned and overwrite it with
   * EINVAL.
   */
  ret = xmlParseURI (escaped_uri);
  if (ret == NULL) {
    errno = EINVAL;
    return NULL;
  }

  return ret;
}
